/* Distributed Checksum Clearinghouse
 *
 * DCC interface daemon
 *
 * Copyright (c) 2003 by Rhyolite Software
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.2.14-1.25 $Revision$
 */

#include "dccif.h"
#include "dcc_ck.h"
#include "dcc_xhdr.h"
#include "dcc_heap_debug.h"
#include "dcc_paths.h"
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#else
#include <sys/pthread.h>
#endif
#include <sys/un.h>
#include <sys/resource.h>
#include <arpa/inet.h>
#include <signal.h>


static int stopping;

static const char *mapfile_nm = DCC_MAP_NM_DEF;

static const char *conn_addr;
static u_char conn_family;
static struct sockaddr_un conn_sun;
static char conn_host[MAXHOSTNAMELEN+1];
static u_int16_t conn_port;
static DCC_SOCKU conn_su;
static const char *raddr_rmask;
static struct in6_addr raddr, rmask;
static int main_sock = -1;

static const char *homedir = 0;
static const char *rundir = DCC_RUNDIR;
static DCC_PATH pidpath;

static const char *tmpdir;

DCC_WF cmn_wf;
static pthread_mutex_t cmn_wf_mutex;

typedef struct dccifd_wf {		/* private white list state */
    struct dccifd_wf *older, *newer;
    pthread_mutex_t mutex;
    DCC_WF	wf;
} DCCIFD_WF;
static pthread_mutex_t dwf_mutex;
static DCCIFD_WF *cur_dwf, wfs[20];

static const char *main_white_nm;
static int main_white_nm_len;

static u_char background = 1;
static u_char try_extra_hard;		/* 1=don't quit if DCC server dead */
static u_char query_only;
static u_char to_white_only;
static u_char grey_on;

static const char *userdirs;
static DCC_PATH userdirs_path;

static struct {
    int	    tgts;			/* total addressees */
    int	    tgts_discarded;		/* discarded for this many addressess */
    int	    tgts_rejected;
    int	    tgts_embargoed;
    time_t  prev;
    time_t  now;
    time_t  next;
} totals;

static enum {
    DCCIFD_SETHDR, DCCIFD_ADDHDR, DCCIFD_NOHDR
} chghdr = DCCIFD_SETHDR;

static u_int dcc_ctxt_sn = 1;

/* per-recipient state */
typedef struct rcpt_st {
    struct rcpt_st *fwd;
    struct work	*wp;
    int		log_fd;
    off_t	log_pos_to;		/* env_To line in main log file */
    off_t	log_pos_white;
    DCC_CKS_WTGTS wtgts;
    DCC_ASK_GREY_RESULT grey_result;
    u_int	embargo_num;
    u_int	flags;
#    define	 RCPT_ST_DISCARD    0x01
#    define	 RCPT_ST_WHITE	    0x02
#    define	 RCPT_ST_BLACK	    0x04
#    define	 RCPT_ST_OK2	    0x08
#    define	 RCPT_ST_GREY	    0x10
#    define	 RCPT_ST_GREY_END   0x20
#    define	 RCPT_ST_GREY_WHITE 0x40
    char	env_to[257];		/* env_to recipient name */
    char	user[257];		/* mailbox */
    DCC_SUM	env_to_sum;
    DCC_SUM	pair_sum;
    DCC_SUM	triple_sum;
    DCC_PATH	dir;			/* recipient's whitelist and logdir */
    DCC_PATH	log_nm;
} RCPT_ST;

/* message state or context */
typedef struct work {
    DCC_CLNT_CTXT *dcc_ctxt;
    u_int	dcc_ctxt_sn;
    int		mta_fd;		/* connection to MTA */
    DCC_SOCKU	mta_addr;
    u_int	xhdr_len;
    char	xhdr[sizeof(DCC_XHDR_START)+sizeof(DCC_BRAND)+1];
    char	clnt_name[MAXHOSTNAMELEN+1];	/* SMTP client */
    char	clnt_str[INET6_ADDRSTRLEN+1];
    struct in6_addr clnt_addr;
    char	helo[DCC_HELO_MAX];
    char	env_from[DCC_HDR_CK_MAX+1];
    DCC_EMSG	emsg;
    DCC_PATH	log_nm;			/* log file for this message */
    RCPT_ST	*rcpt_st, *rcpt_st_last;
    off_t	log_pos_to_first;	/* first env_To line in log file */
    off_t	log_pos_to_end;		/* end of env_To lines in log file */
    off_t	log_pos_white_first;	/* first white-list result */
    off_t	log_pos_white_last;	/* last white-list result */
    int		log_fd;			/* -1=none */
    int		log_fd2;		/* -1=none -2=stop trying */
    int		tmp_fd;
    DCC_PATH	tmp_nm;
    char	sbuf[DCC_HDR_CK_MAX*8];	/* at least DCC_HDR_CK_MAX*2 */
    DCC_HEADER_BUF header;
    /* from here down is zeroed when the structure is allocated */
#define WORK_ZERO fwd
    struct work *fwd;
    char	*sbuf_out;
    char	*sbuf_in;
    char	*sbuf_next_line;
    int		total_hdrs, cr_hdrs;
    u_int	flags;
#    define	 FG_WORK_LOCK	0x0001	/* hold the lock */
#    define	 FG_MTA_ISSPAM	0x0002	/* MTA says it's spam */
#    define	 FG_MTA_BODY	0x0004	/* MTA wants the body */
#    define	 FG_MTA_HEADER	0x0008	/* MTA wants only the X-DCC header */
#    define	 FG_MTA_QUERY	0x0010	/* MTA only wants to query */
#    define	 FG_SEEN_HDR	0x0020	/* have at least 1 header */
#    define	 FG_CK_RCVD	0x0040	/* check 1st Received header */
#    define	 FG_GREY_LIST	0x0080	/* embargo this message */
    int		log_size;
#    define	 MAX_LOG_SIZE	(32*1024)
    DCC_TGTS	tgts, white_tgts, black_tgts, deliver_tgts;
    DCC_TGTS	early_grey_tgts;	/* report to DCC if embargoed */
    DCC_TGTS	late_grey_tgts;		/* don't report to DCC if delivered */
    DCC_GOT_CKS	cks;
    DCC_CKS_WTGTS wtgts;
    u_char	honor;			/* DCC_HONOR_SRVR_ISSPAM etc. */
} WORK;

/* use a free list to avoid malloc() overhead */
static WORK *work_free;

static RCPT_ST *rcpt_st_free;

/* This is a wild guess of open files hidden in libraries and elsewhere.
 * Some systems such as Solaris seem to have an amazing number of them
 * Each whitelist context can involve two open files. */
#define EXTRA_FILES	(32 + DIM(wfs)*2)
/* Every job involves
 *	a socket connected to the MTA,
 *	a log file,
 *	and a socket to talk to the DCC server.
 * While processing per-user whitelists there are
 *	another log file
 *	and an extra file descriptor for the main log file.
 * The file descriptors for the whitelists are accounted for in EXTRA_FILES */
#define FILES_PER_JOB	5


#define MAX_SELECT_WORK ((FD_SETSIZE-EXTRA_FILES)/FILES_PER_JOB)
#define DEF_MAX_WORK	200
#define MIN_MAX_WORK	2
static int max_max_work = MAX_SELECT_WORK;
static int max_work = DEF_MAX_WORK;
static pthread_mutex_t work_mutex;

#define MAX_MTA_DELAY	5		/* max seconds to wait for MTA */


static void sigterm(int);
static void lock_work(void);
static void unlock_work(void);
static void totals_msg(void);
static u_char set_sock(DCC_EMSG, int, const char *);
static void open_unix_sock(void);
static void open_tcp_sock(void);
static void close_main_sock(void);
static void unlink_conn_sun(void);
static void dccifd_error_msg(WORK *, const char *, ...) PATTRIB(2,3);
static void vlog_print(WORK *, const char *p, va_list);
static void log_print(WORK *, const char *, ...) PATTRIB(2,3);
static void *job_start(void *);
static void job_close(WORK *);
static void NRATTRIB job_exit(WORK *);


static void
usage(u_char die)
{
	const char str[] = {
	    "usage: [-VdbxANQGW] [-h homedir]"
	    " [-p /socket/name | host,port]\n"
	    "    [-m map] [-w whiteclnt] [-U userdirs]"
	    " [-t type,[log-thold,][spam-thold]]\n"
	    "    [-g [not-]type] [-S header] [-l logdir] [-R rundir]"
	    " [-T tmpdir]\n"
	    "    [-j maxjobs] [-L ltype,facility.level]"
	};
	static u_char complained;

	/* its important to try to run, so don't give up unless necessary */
	if (die) {
		dcc_logbad(EX_USAGE, complained ? "giving up" : str);
	} else if (!complained) {
		dcc_error_msg("%s\ncontinuing", str);
		complained = 1;
	}
}


int NRATTRIB
main(int argc, char **argv)
{
	DCC_EMSG emsg;
#ifdef HAVE_RLIMIT_NOFILE
	struct rlimit nofile;
	int old_rlim_cur;
#endif
	FILE *f;
	long l;
	u_char log_tgts_set = 0;
	const char *logdir = 0;
	WORK *wp;
	RCPT_ST *rcpt_st;
	char *p;
	time_t now;
	struct timeval delay;
	int secs;
	pthread_t tid;
	int namelen;
	int i, j;

	emsg[0] = '\0';
	dcc_syslog_init(argv[0], 0);
	dcc_init_tholds();

#ifdef HAVE_RLIMIT_NOFILE
	if (0 > getrlimit(RLIMIT_NOFILE, &nofile)) {
		dcc_error_msg("getrlimit(RLIMIT_NOFILE): %s",
			      strerror(errno));
		old_rlim_cur = 1000*1000;
	} else {
		old_rlim_cur = nofile.rlim_cur;
		if (nofile.rlim_max < 1000*1000) {
			i = nofile.rlim_max;
#ifndef USE_POLL
			if (i > FD_SETSIZE)
				i = FD_SETSIZE;
#endif
			j = (i - EXTRA_FILES)/FILES_PER_JOB;
			if (j <= 0) {
				dcc_error_msg("only %d open files allowed",
					      (int)nofile.rlim_max);
			} else {
				max_max_work = j;
			}
		}
	}
#endif /* HAVE_RLIMIT_NOFILE */
	max_work = min(DEF_MAX_WORK, max_max_work);

	while (EOF != (i = getopt(argc, argv,
				  "VdbxANQGWh:p:m:w:U:t:g:S:l:R:T:j:L:"))) {
		switch (i) {
		case 'V':
			fprintf(stderr, DCC_VERSION"\n");
			exit(EX_OK);
			break;

		case 'd':
			++dcc_clnt_debug;
			break;

		case 'b':
			background = 0;
			break;

		case 'x':
			try_extra_hard = 1;
			break;

		case 'A':
			chghdr = DCCIFD_ADDHDR;
			break;

		case 'N':
			chghdr = DCCIFD_NOHDR;
			break;

		case 'Q':
			query_only = 1;
			break;

		case 'G':
			grey_on = 1;
			break;

		case 'W':
			to_white_only = 1;
			break;

		case 'h':
			homedir = optarg;
			break;

		case 'p':
			conn_addr = optarg;
			break;

		case 'm':
			mapfile_nm = optarg;
			break;

		case 'w':
			main_white_nm = optarg;
			main_white_nm_len = strlen(main_white_nm);
			break;

		case 'U':
			if (*optarg == '\0') {
				userdirs = optarg;
			} else {
				snprintf(userdirs_path, sizeof(userdirs_path),
					 "%s/", optarg);
				userdirs = userdirs_path;
			}
			break;

		case 't':
			if (dcc_parse_tholds('t', optarg))
				log_tgts_set = 1;
			break;

		case 'g':		/* honor not-spam "counts" */
			dcc_parse_honor(optarg);
			break;

		case 'S':
			dcc_add_sub_hdr(0, optarg);
			break;

		case 'l':		/* log rejected mail here */
			logdir = optarg;
			break;

		case 'R':
			rundir = optarg;
			break;

		case 'T':
			tmpdir = optarg;
			break;

		case 'j':		/* maximum simultaneous jobs */
			l = strtoul(optarg, &p, 0);
			if (*p != '\0' || l < MIN_MAX_WORK) {
				dcc_error_msg("invalid queue length %s",
					      optarg);
			} else if (l > max_max_work) {
				dcc_error_msg("queue length %s"
					      " larger than limit %d",
					      optarg, max_max_work);
				max_work = max_max_work;
			} else {
				max_work = l;
			}
			break;

		case 'L':
			dcc_parse_log_opt(optarg);
			break;

		default:
			usage(0);
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0)
		usage(0);

	if (!dcc_cdhome(emsg, homedir))
		dcc_error_msg("%s", emsg);

	/* Open the socket before a dccif() function that is starting us
	 * looks for it, and so before our backgrounding fork(). */
	if (!conn_addr) {
		/* use default UNIX domain socket */
		snprintf(conn_sun.sun_path, sizeof(conn_sun.sun_path),
			 "%s/"DCC_DCCIF_UDS, dcc_homedir);
		conn_addr = conn_sun.sun_path;
	}
	p = strchr(conn_addr, ',');
	if (p && strchr(p, ',')) {
		open_tcp_sock();
	} else {
		open_unix_sock();
	}
	if (!set_sock(emsg, main_sock, "main socket"))
		dcc_logbad(dcc_ex_code, "%s", emsg);
	if (0 > listen(main_sock, 10))
		dcc_logbad(EX_IOERR, "listen(): %s", strerror(errno));

	if (logdir)
		dcc_log_init(0, logdir);
	if (!dcc_have_logdir
	    && log_tgts_set)
		dcc_error_msg("log thresholds set with -t but no -l directory");

	if (!tmpdir)
		tmpdir = dcc_have_logdir ? dcc_logdir : _PATH_TMP;

#ifdef HAVE_RLIMIT_NOFILE
	if (old_rlim_cur < (i = max_work*FILES_PER_JOB+EXTRA_FILES)) {
		nofile.rlim_cur = i;
		if (0 > setrlimit(RLIMIT_NOFILE, &nofile)) {
			dcc_error_msg("setrlimit(RLIMIT_NOFILE,%d): %s",
				      i, strerror(errno));
			max_work = old_rlim_cur/FILES_PER_JOB - EXTRA_FILES;
			if (max_work <= 0) {
				dcc_error_msg("only %d open files allowed",
					      old_rlim_cur);
				max_work = MIN_MAX_WORK;
			}
		}
	}
#endif /* HAVE_RLIMIT_NOFILE */

	/* Create the contexts. */
	i = max_work;
	wp = dcc_malloc(sizeof(*wp)*i);
	work_free = wp;
	memset(wp, 0, sizeof(*wp)*i);
	while (--i > 0) {
		wp->mta_fd = -1;
		wp->log_fd = -1;
		wp->log_fd2 = -1;
		wp->tmp_fd = -1;
		wp->fwd = wp+1;
		++wp;
	}

	i = max_work*2;
	rcpt_st = dcc_malloc(sizeof(*rcpt_st)*i);
	rcpt_st_free = rcpt_st;
	memset(rcpt_st, 0, sizeof(*rcpt_st)*i);
	while (--i > 0) {
		rcpt_st->fwd = rcpt_st+1;
		rcpt_st->log_fd = -1;
		++rcpt_st;
	}

	if (background) {
		if (daemon(1, 0) < 0)
			dcc_logbad(EX_OSERR, "daemon(): %s", strerror(errno));
		snprintf(pidpath, sizeof(pidpath), "%s/%s.pid",
			 rundir, dcc_progname);
		unlink(pidpath);
		f = fopen(pidpath, "w");
		if (!f) {
			dcc_error_msg("fopen(%s): %s",
				      pidpath, strerror(errno));
		} else {
			fprintf(f, "%d\n", (u_int)getpid());
			fclose(f);
		}
	}

	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, sigterm);
	signal(SIGTERM, sigterm);
	signal(SIGINT, sigterm);

	/* Be careful to start all threads only after the fork() in daemon(),
	 * because some POSIX threads packages (e.g. FreeBSD) get confused
	 * about threads in the parent.  */

	/* Some pthreads implementations (e.g. AIX) don't like static
	 * POSIX thread initializations */
	pthread_mutex_init(&work_mutex, 0);

	/* start the client library threads and locks */
	dcc_clnt_thread_init(grey_on);
	dcc_wf_init(&cmn_wf, &cmn_wf_mutex, 0);
	for (cur_dwf = wfs; cur_dwf <= LAST(wfs); ++cur_dwf) {
		dcc_wf_init(&cur_dwf->wf, &cur_dwf->mutex, 1);
		cur_dwf->newer = cur_dwf-1;
		cur_dwf->older = cur_dwf+1;
	}
	cur_dwf = wfs;
	LAST(wfs)->older = cur_dwf;
	cur_dwf->newer = LAST(wfs);
	dcc_mutex_init(&dwf_mutex, "");

	if (!dcc_map_info(emsg, mapfile_nm, INVALID_HANDLE_VALUE))
		dcc_logbad(EX_USAGE, "%s", emsg);

	if (main_white_nm) {
		if (!dcc_white_rdy(emsg, &cmn_wf, main_white_nm, 0, 0))
			dcc_error_msg("%s", emsg);
		dcc_wf_unlock(&cmn_wf);
	}

	if (conn_family != AF_UNIX) {
		dcc_trace_msg(DCC_VERSION" listening to %s from %s",
			      conn_addr, raddr_rmask);
	} else {
		dcc_trace_msg(DCC_VERSION" listening to %s",
			      conn_addr);
	}
	if (dcc_clnt_debug)
		dcc_trace_msg("max_work=%d max_max_work=%d",
			      max_work, max_max_work);

	totals.prev  = time(0);
	totals.next = totals.prev+24*60*60;
	while (!stopping) {
		/* do nothing for a second if we are too busy */
		if (work_free == 0) {
			delay.tv_usec = 0;
			delay.tv_sec = 1;
			if (0 > select(0, 0, 0, 0, &delay)
			    && errno != EINTR && errno != EAGAIN)
				dcc_logbad(EX_OSERR, "select():%s",
					   strerror(errno));
			continue;
		}

		/* Recompute delay until we publish totals or quit at
		 * most once per second.  This is cheap enough and deals
		 * with clock changes. */
		now = time(0);
		if (totals.now != now) {
			struct tm tm;

			totals.now = now;
			if (now >= totals.next)
				totals_msg();

			localtime_r(&now, &tm);
			tm.tm_sec = 0;
			tm.tm_min = 0;
			tm.tm_hour = 0;
			++tm.tm_mday;
			totals.next = mktime(&tm);
			if (totals.next == -1) {
				dcc_error_msg("mktime() failed");
				totals.next = now + 60*60;
			}
		}
		secs = totals.next;
		if (secs <= now) {
			secs = 0;
		} else {
			secs -= now;
			/* wake up frequently to notice signals */
			if (secs > 2)
				secs = 2;
		}

		i = dcc_select_poll(emsg, main_sock, 1, secs*DCC_USECS);
		if (i < 0)
			dcc_logbad(EX_OSERR, "%s", emsg);
		if (i == 0)
			continue;

		/* A new connection is ready.  Allocate a context
		 * block and create a thread */
		lock_work();
		wp = work_free;
		if (!wp) {
			/* pretend we weren't listening if we
			 * are out of context blocks */
			unlock_work();
			continue;
		}
		work_free = wp->fwd;
		unlock_work();
		/* clear most of it */
		memset(&wp->WORK_ZERO, 0,
		       sizeof(*wp)
		       - ((char*)&wp->WORK_ZERO - (char*)wp));
		wp->emsg[0] = '\0';

		namelen = sizeof(wp->mta_addr.sa);
		wp->mta_fd = accept(main_sock, &wp->mta_addr.sa,
				    &namelen);
		if (wp->mta_fd < 0) {
			dcc_error_msg("accept(): %s", strerror(errno));
			job_close(wp);
			continue;
		}
		if (conn_family != AF_UNIX) {
			struct in6_addr addr6, *ap;

			if (wp->mta_addr.sa.sa_family == AF_INET6) {
				ap = &wp->mta_addr.ipv6.sin6_addr;
			} else {
				ap = &addr6;
				dcc_toipv6(ap, wp->mta_addr.ipv4.sin_addr);
			}
			if ((ap->s6_addr32[0]
			     & rmask.s6_addr32[0]) != raddr.s6_addr32[0]
			    || (ap->s6_addr32[1]
				& rmask.s6_addr32[1]) != raddr.s6_addr32[1]
			    || (ap->s6_addr32[2]
				& rmask.s6_addr32[2]) != raddr.s6_addr32[2]
			    || (ap->s6_addr32[3]
				& rmask.s6_addr32[3]) != raddr.s6_addr32[3]) {
				dcc_error_msg("unauthorized client address %s",
					      dcc_su2str(&wp->mta_addr));
				job_close(wp);
				continue;
			}
		}
		if (!set_sock(emsg, wp->mta_fd, "MTA")) {
			dcc_error_msg("%s", emsg);
			job_close(wp);
			continue;
		}
		pthread_create(&tid, 0, job_start, wp);
		i = pthread_detach(tid);
		if (i)
			dcc_logbad(EX_SOFTWARE, "pthread_detach(): %s",
				   strerror(i));
	}
	totals_msg();

	close_main_sock();
	if (pidpath[0] != '\0')
		unlink(pidpath);

	exit(-stopping);
}



static void
lock_work(void)
{
	int result = pthread_mutex_lock(&work_mutex);
	if (result)
		dcc_logbad(EX_SOFTWARE, "pthread_mutex_lock(work_free): %s",
			   strerror(result));
}



static void
unlock_work(void)
{
	int result = pthread_mutex_unlock(&work_mutex);
	if (result)
		dcc_logbad(EX_SOFTWARE, "pthread_mutex_unlock(work_free): %s",
			   strerror(result));
}



/* brag about our accomplishments */
static void
totals_msg(void)
{
	struct timeval now;
	struct tm tm;
	char tbuf[20];

	lock_work();
	gettimeofday(&now, 0);
	strftime(tbuf, sizeof(tbuf), "%x %X",
		 localtime_r(&totals.prev, &tm));
	if (grey_on) {
		dcc_trace_msg(DCC_VERSION
			      " greylisted %d messages,"
			      " rejected messages to %d targets and"
			      " discarded messages to %d targets among"
			      " %d total since %s",
			      totals.tgts_embargoed,
			      totals.tgts_rejected,
			      totals.tgts_discarded,
			      totals.tgts, tbuf);
	} else {
		dcc_trace_msg(DCC_VERSION
			      " rejected messages to %d targets and"
			      " discarded messages to %d targets among"
			      " %d total since %s",
			      totals.tgts_rejected,
			      totals.tgts_discarded,
			      totals.tgts, tbuf);
	}
	memset(&totals, 0, sizeof(totals));
	totals.prev = now.tv_sec;
	unlock_work();
}



static void
unlink_conn_sun(void)
{
	if (conn_family != AF_UNIX)
		return;

	/* there is an unavoidable race here */
	if (0 > unlink(conn_sun.sun_path)
	    && errno != ENOENT)
		dcc_error_msg("unlink(%s): %s",
			      conn_sun.sun_path, strerror(errno));
}



static void
open_unix_sock(void)
{
	DCC_EMSG emsg;
	struct stat sb;

	emsg[0] = '\0';
	conn_family = AF_UNIX;

	if (conn_addr != conn_sun.sun_path) {
		if (strlen(optarg) >= ISZ(conn_sun.sun_path))
			dcc_logbad(EX_USAGE, "invalid UNIX domain socket: %s",
				   conn_addr);
		strcpy(conn_sun.sun_path, optarg);
	}
#ifdef DCC_HAVE_SA_LEN
	conn_sun.sun_len = SUN_LEN(&conn_sun);
#endif
	conn_sun.sun_family = AF_UNIX;

	if (0 <= stat(conn_sun.sun_path, &sb)
	    && !(S_ISSOCK(sb.st_mode) || S_ISFIFO(sb.st_mode)))
		dcc_logbad(EX_UNAVAILABLE, "non-socket present at %s",
			   conn_sun.sun_path);

	/* look for a daemon already using our socket */
	main_sock = socket(AF_UNIX, SOCK_STREAM, 0);
	if (main_sock < 0)
		dcc_logbad(EX_OSERR, "socket(AF_UNIX): %s", strerror(errno));
	/* unlink it only if it looks like a dead socket */
	if (0 > connect(main_sock, (struct sockaddr *)&conn_sun,
			sizeof(conn_sun))) {
		if (errno == ECONNREFUSED || errno == ECONNRESET) {
			unlink_conn_sun();
		} else if (dcc_clnt_debug > 1) {
			dcc_error_msg("connect(old server %s): %s",
				      conn_sun.sun_path, strerror(errno));
		}
	} else {
		/* connect() worked so the socket is alive */
		dcc_logbad(EX_UNAVAILABLE,
			   "something already or still running with socket at"
			   " %s",
			   conn_sun.sun_path);
	}
	close(main_sock);

	main_sock = socket(AF_UNIX, SOCK_STREAM, 0);
	if (main_sock < 0)
		dcc_logbad(EX_OSERR, "socket(AF_UNIX): %s", strerror(errno));
	if (0 > bind(main_sock, (struct sockaddr *)&conn_sun,
		     sizeof(conn_sun)))
		dcc_logbad(EX_IOERR, "bind(%s) %s",
			   conn_sun.sun_path, strerror(errno));
	if (0 >	chmod(conn_sun.sun_path, 0666))
		dcc_error_msg("chmod(%s, 0666): %s",
			      conn_sun.sun_path, strerror(errno));
}



static void
open_tcp_sock(void)
{
	DCC_EMSG emsg;
	const char *cp;
	const struct hostent *hp;
	char *rhost;
	int error;

	emsg[0] = '\0';

	rhost = strchr(conn_addr, ',');
	if (!rhost)
		dcc_logbad(EX_USAGE, "missing port number: \"%s\"",
			   conn_addr);
	rhost = strchr(rhost+1, ',');
	if (!rhost) {
		dcc_logbad(EX_USAGE, "missing remote host name: \"%s\"",
			   conn_addr);
	}
	*rhost++ = '\0';
	error = dcc_str2cidr(emsg, &raddr, &rmask, rhost, 0, 0);
	if (error <= 0) {
		dcc_logbad(EX_USAGE, "invalid remote host and mask: \"%s\"",
			   rhost);
	}
	raddr_rmask = rhost;

	cp = dcc_parse_nm_port(emsg, conn_addr, DCC_GET_PORT_INVALID,
			       conn_host, sizeof(conn_host),
			       &conn_port, 0, 0,
			       0, 0);
	if (!cp)
		dcc_logbad(dcc_ex_code, "%s", emsg);
	if (*cp != '\0')
		dcc_logbad(EX_USAGE, "invalid IP address: \"%s\"", conn_addr);

	if (conn_host[0] == '\0' || !strcmp(conn_host, "@")) {
			/* null or "@" means INADDR_ANY */
		dcc_mk_su(&conn_su, AF_INET, 0, conn_port);
	} else {
#ifdef NO_IPV6
		hp = dcc_get_host(conn_host, 0, &error);
#else
		hp = dcc_get_host(conn_host, 2, &error);
#endif
		if (!hp)
			dcc_logbad(EX_NOHOST, "%s: %s",
				   conn_host, hstrerror(error));
		dcc_mk_su(&conn_su, hp->h_addrtype, hp->h_addr, conn_port);
		dcc_freehostent();
	}

	main_sock = socket(conn_su.sa.sa_family, SOCK_STREAM, 0);
	if (main_sock < 0)
		dcc_logbad(EX_OSERR, "socket(): %s", strerror(errno));
	if (0 > bind(main_sock, &conn_su.sa, DCC_SU_LEN(&conn_su)))
		dcc_logbad(EX_UNAVAILABLE, "bind(%s) %s",
			   conn_addr, strerror(errno));

	conn_family = conn_su.sa.sa_family;
}



static u_char
set_sock(DCC_EMSG emsg, int s, const char *sname)
{
	int on;

	if (0 > fcntl(s, F_SETFD, FD_CLOEXEC)) {
		dcc_pemsg(EX_IOERR, emsg,
			  "fcntl(%s, F_SETFD, FD_CLOEXEC): %s",
			  sname, strerror(errno));
		return 0;
	}

	if (conn_family != AF_UNIX) {
		on = 1;
		if (0 > setsockopt(s, SOL_SOCKET, SO_KEEPALIVE,
				   &on, sizeof(on))) {
			dcc_pemsg(EX_IOERR, emsg,
				  "setsockopt(%s, SO_KEEPALIVE): %s",
				  sname, strerror(errno));
			return 0;
		}
	}

	if (-1 == fcntl(s, F_SETFL,
			fcntl(s, F_GETFL, 0) | O_NONBLOCK)) {
		dcc_pemsg(EX_OSERR, emsg, "fcntl(%s, O_NONBLOCK): %s",
			  sname, strerror(errno));
		return 0;
	}

	return 1;
}



/* close a socket to the MTA */
static void
sock_close(WORK *wp)
{
	if (wp->mta_fd >= 0) {
		if (0 > close(wp->mta_fd))
			dcc_error_msg("close(socket): %s", strerror(errno));
		wp->mta_fd = -1;
	}
}



static u_char				/* 0=EOF, 1=read something */
sock_read(WORK *wp)
{
	int fspace, len, total, i;

	if (wp->sbuf_out >= wp->sbuf_in)
		wp->sbuf_out = wp->sbuf_in = wp->sbuf;

	fspace = &wp->sbuf[sizeof(wp->sbuf)] - wp->sbuf_in;
	if (fspace < DCC_HDR_CK_MAX) {
		if (wp->sbuf_out == wp->sbuf) {
			dccifd_error_msg(wp, "buffer overrun; in=%d",
					 wp->sbuf_in - wp->sbuf);
			job_exit(wp);
		}
		len = wp->sbuf_in - wp->sbuf_out;
		memmove(wp->sbuf, wp->sbuf_out, len);
		wp->sbuf_out = wp->sbuf;
		wp->sbuf_in = wp->sbuf_out+len;
		fspace = sizeof(wp->sbuf) - len;
	}

	if (wp->flags & FG_WORK_LOCK)
		unlock_work();

	for (;;) {
		i = dcc_select_poll(wp->emsg, wp->mta_fd, 1,
				    MAX_MTA_DELAY*DCC_USECS);
		if (i > 0)
			break;
		if (i < 0) {
			dccifd_error_msg(wp, "%s", wp->emsg);
		} else {
			dccifd_error_msg(wp, "MTA read timeout");
		}
		job_exit(wp);
	}
	total = read(wp->mta_fd, wp->sbuf_in, fspace);
	if (total < 0) {
		dccifd_error_msg(wp, "read(sock): %s", strerror(errno));
		job_exit(wp);
	}
	wp->sbuf_in += total;

	if (wp->flags & FG_WORK_LOCK)
		lock_work();
	return total != 0;
}



/* ensure there is another line in wp->sbuf and add a terminal '\0' */
static int				/* bytes available */
sock_read_line(WORK *wp)
{
	int i;
	char *p;

	for (;;) {
		if (0 != (i = wp->sbuf_in - wp->sbuf_out)
		    && 0 != (p = memchr(wp->sbuf_out, '\n', i))) {
			wp->sbuf_next_line = p+1;
			for (;;) {
				*p-- = '\0';
				if (p < wp->sbuf_out)
					return 0;
				if (*p != '\r')
					return p+1 - wp->sbuf_out;
			}
		}

		if (!sock_read(wp)) {
			dccifd_error_msg(wp, "truncated request");
			job_exit(wp);
		}
	}
}



/* Look for a string from the MTA */
static u_char
sock_ck_str(WORK *wp, const char *st, int stlen)
{
	while (wp->sbuf_out < wp->sbuf_next_line
	       && (*wp->sbuf_out == '\t' || *wp->sbuf_out == ' '))
		++wp->sbuf_out;

	if (stlen <= wp->sbuf_next_line - wp->sbuf_out
	    && !strncasecmp(wp->sbuf_out, st, stlen)) {
		if (wp->sbuf_out[stlen] == '\0') {
			wp->sbuf_out += stlen;
			return 1;
		}
		if (wp->sbuf_out[stlen] == '\t'
		    || wp->sbuf_out[stlen] == ' ') {
			do {
				++stlen;
			} while (wp->sbuf_out[stlen] == '\t'
				 || wp->sbuf_out[stlen] == ' ');
			wp->sbuf_out += stlen;
			return 1;
		}
	}

	return 0;
}



static void
sock_write(WORK *wp, const void *buf, int len)
{
	int i;

	if (wp->mta_fd < 0)
		dcc_logbad(EX_SOFTWARE, "attempt to write closed socket");

	do {
		i = dcc_select_poll(wp->emsg, wp->mta_fd, 0,
				    MAX_MTA_DELAY*DCC_USECS);
		if (i < 0) {
			dccifd_error_msg(wp, "%s", wp->emsg);
			job_exit(wp);
		}
		if (i == 0) {
			dccifd_error_msg(wp, "MTA write timeout");
			job_exit(wp);
		}
		i = write(wp->mta_fd, buf, len);
		if (i < 0) {
			if (errno == EWOULDBLOCK || errno == EAGAIN)
				continue;
			dccifd_error_msg(wp, "write(MTA socket,%d): %s",
					 len, strerror(errno));
			job_exit(wp);
		}
		if (i == 0) {
			dccifd_error_msg(wp, "write(MTA socket,%d)=%d",
					 len, i);
			job_exit(wp);
		}
		len -= i;
	} while (len > 0);
}



static void
tmp_write(WORK *wp, const void *buf, int len)
{
	int i;

	if (wp->tmp_fd < 0)
		return;

	i = write(wp->tmp_fd, buf, len);
	if (i != len) {
		if (i < 0)
			dccifd_error_msg(wp, "write(%s,%d): %s",
					 wp->tmp_nm, len, strerror(errno));
		else
			dccifd_error_msg(wp, "write(%s,%d)=%d",
					 wp->tmp_nm, len, i);
		job_exit(wp);
	}
}



static int
tmp_read(WORK *wp, void *buf, int len)
{
	int i;

	if (wp->tmp_fd < 0)
		return 0;

	i = read(wp->tmp_fd, buf, len);
	if (i < 0) {
		dccifd_error_msg(wp, "read(%s,%d): %s",
				 wp->tmp_nm, len, strerror(errno));
		job_exit(wp);
	}
	return i;
}



/* fill wp->sbuf */
static int
tmp_read_sbuf(WORK *wp)
{
	int i;

	/* preserving what is already in it */
	i = wp->sbuf_in - wp->sbuf_out;
	if (i > 0)
		memmove(wp->sbuf, wp->sbuf_out, i);
	wp->sbuf_out = wp->sbuf;
	wp->sbuf_in = wp->sbuf_out+i;

	i = tmp_read(wp, wp->sbuf_in, ISZ(wp->sbuf) - i);
	if (i == 0)
		return 0;
	wp->sbuf_in += i;
	return 1;
}



/* start log file */
static void
log_start(WORK *wp)
{
	time_t now;
	struct tm tm;
	char date_buf[40];

	if (!dcc_have_logdir)
		return;
	if (wp->log_fd >= 0)
		return;

	wp->log_size = 0;
	wp->log_fd = dcc_log_open(0, wp->log_nm, sizeof(wp->log_nm));
	if (wp->log_fd < 0)
		return;

	now = time(0);
	strftime(date_buf, sizeof(date_buf), DCC_LOG_DATE_FMT,
		 localtime_r(&now, &tm));
	log_print(wp, DCC_LOG_DATE_PAT"\n", date_buf);
}



static void
log_stop(WORK *wp)
{
	if (wp->log_fd2 >= 0
	    && 0 > close(wp->log_fd2))
		dcc_error_msg("close(%s): %s", wp->log_nm, strerror(errno));
	wp->log_fd2 = -1;

	if (wp->log_fd < 0)
		return;

	if (!(wp->honor & DCC_HONOR_LOGIT)) {
		/* Delete the log file if it is not interesting */
		unlink(wp->log_nm);
	} else {
		/* give it a permanent name if it is interesting */
		dcc_log_keep(0, wp->log_nm, sizeof(wp->log_nm));
	}
	if (0 > close(wp->log_fd))
		dcc_error_msg("close(%s): %s", wp->log_nm, strerror(errno));
	wp->log_fd = -1;
}



static void
log_write(WORK *wp, const void *logbuf, u_int len)
{
	int result;

	if (wp->log_fd < 0)
		return;

	if (!len)
		len = strlen(logbuf);
	wp->log_size += len;

	result = write(wp->log_fd, logbuf, len);
	if ((int)len == result)
		return;

	if (result < 0)
		dcc_error_msg("write(%s): %s", wp->log_nm, strerror(errno));
	else
		dcc_error_msg("write(%s,%d)=%d", wp->log_nm, len, result);
	job_exit(wp);
}



static void
log_ck_write(void *wp, const void *logbuf, u_int len)
{
	log_write((WORK *)wp, logbuf, len);
}



#define LOG_CAPTION(wp, s) log_write((wp), (s), sizeof(s)-1)
#define LOG_EOL(wp) LOG_CAPTION((wp), "\n")



/* put something into a log file */
static void
vlog_print(WORK *wp, const char *p, va_list args)
{
	char logbuf[MAXHOSTNAMELEN*2];

	if (wp->log_fd < 0)
		return;

	vsnprintf(logbuf, sizeof(logbuf), p, args);
	log_write(wp, logbuf, 0);
}



static void
log_print(WORK *wp, const char *p, ...)
{
	va_list args;

	va_start(args, p);
	vlog_print(wp, p, args);
	va_end(args);
}



static off_t
log_lseek(WORK *wp, int whence)
{
	off_t result;

	if (wp->log_fd < 0)
		return 0;
	result = lseek(wp->log_fd, 0, whence);
	if (result == -1) {
		dccifd_error_msg(wp, "lseek(%s): %s",
				 wp->log_nm, strerror(errno));
		job_exit(wp);
	}
	return result;
}



static void
dccifd_error_msg(WORK *wp, const char *p, ...)
{
	va_list args;

	va_start(args, p);
	dcc_verror_msg(p, args);
	vlog_print(wp, p, args);
	va_end(args);
	LOG_EOL(wp);
	wp->honor |= DCC_HONOR_LOGIT;
}



static void
dccifd_trace_msg(WORK *wp, const char *p, ...)
{
	va_list args;

	va_start(args, p);
	dcc_vtrace_msg(p, args);
	vlog_print(wp, p, args);
	va_end(args);
	LOG_EOL(wp);
	wp->honor |= DCC_HONOR_LOGIT;
}



static void
job_close(WORK *wp)
{
	RCPT_ST *rcpt_st, *next_rcpt_st;

	if (wp->tmp_fd >= 0) {
		if (0 > close(wp->tmp_fd))
			dccifd_error_msg(wp, "close(%s): %s",
					 wp->tmp_nm, strerror(errno));
		wp->tmp_fd = -1;
	}
	sock_close(wp);
	log_stop(wp);

	if (!(wp->flags & FG_WORK_LOCK))
		lock_work();
	rcpt_st = wp->rcpt_st;
	if (rcpt_st) {
		wp->rcpt_st = 0;
		do {
			next_rcpt_st = rcpt_st->fwd;
			if (rcpt_st->log_fd >= 0) {
				if (0 > close(rcpt_st->log_fd))
					dccifd_error_msg(wp, "close(%s): %s",
							rcpt_st->log_nm,
							strerror(errno));
				rcpt_st->log_fd = -1;
			}
			rcpt_st->fwd = rcpt_st_free;
			rcpt_st_free = rcpt_st;
		} while ((rcpt_st = next_rcpt_st) != 0);
	}

	wp->fwd = work_free;
	work_free = wp;
	unlock_work();
	wp->flags &= ~FG_WORK_LOCK;
}



static void NRATTRIB
job_exit(WORK *wp)
{
	job_close(wp);
	pthread_exit(0);
	/* mostly to suppress warning */
	dcc_logbad(EX_SOFTWARE, "pthread_exit() returned");
}



/* write headers lines from the buffer */
static void
hdr_write(WORK *wp, char *bol)
{
	int i;

	i = bol - wp->sbuf_out;
	if (i > 0) {
		sock_write(wp, wp->sbuf_out, i);
		wp->sbuf_out = bol;
	}
}



static void
hdrs_copy(WORK *wp)
{
	enum {START_HDR, SKIP_HDR, COPY_HDR} cmode;
	char *bol, *nl, *eol;
	int i;

	if (-1 == lseek(wp->tmp_fd, 0, SEEK_SET)) {
		dccifd_error_msg(wp, "rewind %s: %s",
				 wp->tmp_nm, strerror(errno));
		job_exit(wp);
	}

	cmode = START_HDR;
	wp->sbuf_in = wp->sbuf_out = wp->sbuf;
	for (;;) {
		/* fill wp->sbuf while keeping anything already present */
		if (!tmp_read_sbuf(wp))
			return;

		bol = wp->sbuf_out;
		while ((i = wp->sbuf_in - bol) > 0) {
			/* Find the end of the next line. */
			eol = memchr(bol, '\n', i);
			if (!eol) {
				/* Fill the buffer if we can't find '\n''
				 * and the buffer has room */
				if (i < ISZ(wp->sbuf))
					break;
				eol = wp->sbuf_in-1;
				nl = 0;
			} else if (eol+1 >= wp->sbuf_in) {
				/* insist on having the character after '\n'
				 * or pretend we could not find it */
				if (i < ISZ(wp->sbuf))
					break;
				--eol;
				nl = 0;
			} else {
				nl = eol+1;
			}

			if (cmode == START_HDR) {
				/* We are at start of a header or the body.
				 * Quit before line of "\n" or "\r\n" */
				if (eol == bol
				    || (eol == bol+1 && *bol == '\r')) {
					/* write any preceding lines */
					hdr_write(wp, bol);
					return;
				}

				/* Look for our header
				 * Assume the buffer is larger than the
				 * largest possible X-DCC header. */
				if (bol[wp->xhdr_len] == ':'
				    && !strncasecmp(bol, wp->xhdr, wp->xhdr_len)
				    && chghdr == DCCIFD_SETHDR) {
					/* Found it and we want to skip it.
					 * Copy any preceding headers. */
					hdr_write(wp, bol);
					cmode = SKIP_HDR;
				} else {
					cmode = COPY_HDR;
				}
			}
			if (cmode == SKIP_HDR)
				wp->sbuf_out = eol+1;

			/* Check the character after '\n' for
			 * whitespace indicating continuation */
			if (nl && *nl  != ' ' && *nl != '\t')
				cmode = START_HDR;
			bol = eol+1;
		}

		hdr_write(wp, bol);
	}
}



/* we are finished with one SMTP message.
 * get ready for the next from the same connection to an SMTP client */
static void NRATTRIB
job_done(WORK *wp, char result_char, const char *result_str)
{
	RCPT_ST *rcpt_st;
	char *p;
	int i;

	LOG_CAPTION(wp, "result: ");
	log_write(wp, result_str, 0);
	LOG_EOL(wp);

	/* tell MTA overall result */
	p = wp->sbuf;
	*p++ = result_char;
	*p++ = '\n';

	/* output list of recipients */
	for (rcpt_st = wp->rcpt_st; rcpt_st; rcpt_st = rcpt_st->fwd) {
		if (p >= &wp->sbuf[sizeof(wp->sbuf)-1]) {
			sock_write(wp, wp->sbuf, sizeof(wp->sbuf)-1);
			p = wp->sbuf;
		}
		if (result_char == DCCIF_RESULT_GREY) {
			*p++ = DCCIF_RCPT_GREY;
		} else if (rcpt_st->flags & RCPT_ST_DISCARD) {
			*p++ = DCCIF_RCPT_REJECT;
			/* tell greylist to restore embargo for those targets
			 * that believe the message was spam and did not
			 * white- or blacklist it */
			if (rcpt_st->grey_result != DCC_GREY_OFF
			    && !dcc_grey_spam(wp->emsg, wp->dcc_ctxt,
					      rcpt_st->triple_sum))
				dccifd_error_msg(wp, "%s",wp->emsg);
		} else {
			*p++ = DCCIF_RCPT_ACCEPT;
		}
	}
	*p++ = '\n';
	sock_write(wp, wp->sbuf, p-wp->sbuf);

	if (wp->flags & FG_MTA_BODY) {
		hdrs_copy(wp);
		if (wp->header.buf[0] != '\0'
		    && chghdr != DCCIFD_NOHDR) {
			/* write X-DCC */
			sock_write(wp, wp->header.buf, wp->header.used);
			if (wp->cr_hdrs > wp->total_hdrs/2) {
				/* end the X-DCC header with "\r\n" if at least
				 * half of the header lines ended that way */
				sock_write(wp, "\r\n", 2);
			} else {
				/* otherwise use only "\n" */
				sock_write(wp, "\n", 1);
			}
		}
		do {			/* copy body */
			i = wp->sbuf_in - wp->sbuf_out;
			if (i > 0) {
				sock_write(wp, wp->sbuf_out, i);
				wp->sbuf_out = wp->sbuf_in;
			}
		} while (tmp_read_sbuf(wp));

	} else if (wp->flags & FG_MTA_HEADER) {
		/* MTA wants only the header, if we have it */
		if (wp->header.used != 0)
			sock_write(wp, wp->header.buf, wp->header.used);
		sock_write(wp, "\n", 1);
	}

	job_exit(wp);
}



static void
close_main_sock(void)
{
	if (main_sock < 0)
		return;

	unlink_conn_sun();
	if (0 > close(main_sock))
		dcc_error_msg("close(main socket): %s",
			      strerror(errno));
	main_sock = -1;
}



/* watch for fatal signals */
static void
sigterm(int sig)
{
	stopping = sig;
	unlink_conn_sun();
	dcc_clnt_stop_resolve();
}



/* find and lock a DCC_WF */
static DCC_WF*
find_wf(RCPT_ST *rcpt_st)
{
	DCCIFD_WF *dwf;
	DCC_PATH white_nm;
	const char *white_nm_ptr;
	struct stat sb;
	int i;

	snprintf(white_nm, sizeof(white_nm), "%s/whiteclnt", rcpt_st->dir);
	if (0 > stat(white_nm, &sb)) {
		if (errno != ENOENT
		    && (errno != ENOTDIR || dcc_clnt_debug))
			dccifd_error_msg(rcpt_st->wp, "stat(%s): %s",
				       white_nm, strerror(errno));
		return 0;
	}

	i = pthread_mutex_lock(&dwf_mutex);
	if (i)
		dcc_logbad(EX_SOFTWARE, "pthread_mutex_lock(dwf_mutex): %s",
			   strerror(i));

	dwf = cur_dwf;
	for (;;) {
		dcc_wf_lock(&dwf->wf);
		if (dwf->older == cur_dwf) {
			white_nm_ptr = white_nm;
			break;
		}

		if (!strcmp(white_nm, dwf->wf.ascii_nm)) {
			white_nm_ptr = 0;
			break;
		}

		dcc_wf_unlock(&dwf->wf);
		dwf = dwf->older;
	}

	/* move to front */
	if (dwf != cur_dwf) {
		dwf->older->newer = dwf->newer;
		dwf->newer->older = dwf->older;
		dwf->older = cur_dwf;
		dwf->newer = cur_dwf->newer;
		dwf->newer->older = dwf;
		dwf->older->newer = dwf;
		cur_dwf = dwf;
	}
	i = pthread_mutex_unlock(&dwf_mutex);
	if (i)
		dcc_logbad(EX_SOFTWARE, "pthread_mutex_unlock(dwf_mutex): %s",
			   strerror(i));

	if (!dcc_white_rdy(rcpt_st->wp->emsg, &dwf->wf, white_nm_ptr, 2, 0)) {
		dcc_wf_unlock(&dwf->wf);
		dccifd_error_msg(rcpt_st->wp, "%s",
			       rcpt_st->wp->emsg);
		return 0;
	}

	return &dwf->wf;
}



/* check a white list for a target address */
static u_char				/* 1=black or white result */
ask_white_to(WORK *wp, RCPT_ST *rcpt_st, DCC_WF *wf, DCC_SUM sum,
	     const char *str,
	     u_char wf_locked,		/* 1=mutex already held */
	     u_char keep_lock)		/* 0=unlock both if unambiguous */
{
	switch (dcc_white_sum(wp->emsg, wf,
			      DCC_CK_ENV_TO, sum, wf_locked)) {
	case DCC_WHITE_ERROR:
		dccifd_error_msg(wp, "%s", wp->emsg);
		rcpt_st->flags |= RCPT_ST_WHITE;    /* fail conservatively */
		break;

	case DCC_WHITE_HALF_LISTED:
		/* "OK2" for the RCPT in the local white list does not mean
		 * the address is half white listed, but that it is
		 * explicitly ok to reject or discard spam for it. */
		rcpt_st->flags |= RCPT_ST_OK2;
		return 0;

	case DCC_WHITE_LISTED:
		log_print(wp, "%s(env_To %s)-->OK\n",
			  DCC_NM2PATH(wf->ascii_nm), str);
		rcpt_st->flags |= RCPT_ST_WHITE;
		break;

	case DCC_WHITE_UNLISTED:
		return 0;

	case DCC_WHITE_BLACK:
		log_print(wp, "%s(env_To %s)-->spam\n",
			  DCC_NM2PATH(wf->ascii_nm), str);
		rcpt_st->flags |= RCPT_ST_BLACK;
		break;
	}

	/* we have an umabiguous answer */
	if (!keep_lock) {
		if (!dcc_white_unlock(wp->emsg, wf, 0)) {
			dccifd_error_msg(wp, "%s", wp->emsg);
			/* fail conservatively */
			rcpt_st->flags |= RCPT_ST_WHITE;
		}
		dcc_wf_unlock(wf);
	}
	return 1;
}



static void
log_white_msg(WORK *wp, DCC_WF *wf, DCC_WHITE_RESULT result)
{
	switch (result) {
	case DCC_WHITE_ERROR:
	case DCC_WHITE_LISTED:
		log_print(wp, "%s-->OK\n", DCC_NM2PATH(wf->ascii_nm));
		break;

	case DCC_WHITE_HALF_LISTED:
	case DCC_WHITE_UNLISTED:
		break;

	case DCC_WHITE_BLACK:
		log_print(wp, "%s-->spam\n", DCC_NM2PATH(wf->ascii_nm));
		break;
	}
}



/* check a white list for the checksums of the message */
static inline u_char			/* 1=black or white result */
ck_white_msg(RCPT_ST *rcpt_st, DCC_WHITE_RESULT result)
{
	switch (result) {
	case DCC_WHITE_ERROR:		/* fail conservatively */
	case DCC_WHITE_LISTED:
		rcpt_st->flags |= RCPT_ST_WHITE;
		break;

	case DCC_WHITE_HALF_LISTED:
	case DCC_WHITE_UNLISTED:
		return 0;

	case DCC_WHITE_BLACK:
		rcpt_st->flags |= RCPT_ST_BLACK;
		break;
	}
	return 1;
}



/* check the white lists for one target */
static void
ask_white_rcpt(WORK *wp, RCPT_ST *rcpt_st, DCC_WHITE_RESULT global_result)
{
	DCC_SUM addr_sum;
	DCC_WHITE_RESULT private_result;
	DCC_WF *wf;

	rcpt_st->log_pos_white = log_lseek(wp, SEEK_END);

	if (wp->flags & FG_MTA_ISSPAM) {
		rcpt_st->flags |= RCPT_ST_BLACK;
		return;
	}

	dcc_str2ck(rcpt_st->env_to_sum, 0, 0, rcpt_st->env_to);
	if (rcpt_st->dir[0] == '\0') {
		wf = 0;
	} else {
		if (rcpt_st->user[0])
			dcc_str2ck(addr_sum, 0, 0, rcpt_st->user);

		/* check env_to & mailbox name in private white list */
		wf = find_wf(rcpt_st);
		if (wf) {
			if (ask_white_to(wp, rcpt_st, wf, rcpt_st->env_to_sum,
					 rcpt_st->env_to, 1, 0))
				return;
			/* then check the mailbox name */
			if (rcpt_st->user[0]
			    && ask_white_to(wp, rcpt_st, wf, addr_sum,
					    rcpt_st->user, 1, 0))
				return;
		}
	}

	/* check global white list for env_to & mailbox */
	if (ask_white_to(wp, rcpt_st, &cmn_wf, rcpt_st->env_to_sum,
			 rcpt_st->env_to, 1, 1))
		return;
	if (rcpt_st->user[0]
	    && ask_white_to(wp, rcpt_st, &cmn_wf, addr_sum,
			    rcpt_st->user, 1, 1))
		return;

	/* check other checksums in private white list */
	if (wf) {
		private_result = dcc_white_cks(wp->emsg, wf, &wp->cks,
					       rcpt_st->wtgts, 1, 0);
		if (ck_white_msg(rcpt_st, private_result)) {
			log_white_msg(rcpt_st->wp, wf, private_result);
			return;
		}
	}

	/* check other checksums in global white list */
	if (ck_white_msg(rcpt_st, global_result))
		memcpy(rcpt_st->wtgts, wp->wtgts,
		       sizeof(rcpt_st->wtgts));
}



/* check the white lists for all targets */
static void
ask_white(WORK *wp)
{
	RCPT_ST *rcpt_st;
	DCC_WHITE_RESULT global_result;
	u_char local_grey_on;

	/* Check the main white list and the white lists for the recipients.
	 * Use the main whitelist only for recipients whose individual
	 * white lists don't give a black or white answer.
	 * Check the main white list first (and so even if not necessary)
	 * so that problems with it are in all of the logs. */
	global_result = dcc_white_cks(wp->emsg, &cmn_wf, &wp->cks, wp->wtgts,
				      0, 1);
	if (global_result == DCC_WHITE_ERROR) {
		dccifd_error_msg(wp, "%s", wp->emsg);
		global_result = DCC_WHITE_LISTED;   /* fail conservatively */
	}
	log_white_msg(wp, &cmn_wf, global_result);

	wp->log_pos_white_first = log_lseek(wp, SEEK_END);

	local_grey_on = grey_on;
	for (rcpt_st = wp->rcpt_st; rcpt_st; rcpt_st = rcpt_st->fwd) {
		rcpt_st->grey_result = DCC_GREY_OFF;
		ask_white_rcpt(wp, rcpt_st, global_result);
		if (rcpt_st->flags & RCPT_ST_BLACK) {
			wp->honor |= (DCC_HONOR_LOGIT | DCC_HONOR_LOCAL_ISSPAM);

		} else if (rcpt_st->flags & RCPT_ST_WHITE) {
			++wp->white_tgts;
			++wp->deliver_tgts;

		} else if (local_grey_on) {
			rcpt_st->grey_result = dcc_ask_grey(wp->emsg,
							wp->dcc_ctxt,
							query_only
							? DCC_OP_GREY_QUERY
							: DCC_OP_GREY_REPORT,
							rcpt_st->pair_sum,
							rcpt_st->triple_sum,
							&wp->cks,
							rcpt_st->env_to_sum,
							&rcpt_st->embargo_num,
							&wp->early_grey_tgts,
							&wp->late_grey_tgts,
							&wp->honor);
			switch (rcpt_st->grey_result) {
			case DCC_GREY_OFF:
				break;
			case DCC_GREY_ASK_FAIL:
				dccifd_error_msg(wp, "%s", wp->emsg);
				/* If we are already going to embargo the
				 * message, assume this would have been
				 * the same.
				 * If we are trying hard, assume the
				 * message would have been embargoed */
				if ((wp->flags & FG_GREY_LIST)
				    || try_extra_hard) {
					rcpt_st->flags |= RCPT_ST_GREY;
					wp->flags |= FG_GREY_LIST;
					break;
				}
				/* otherwise give up on greylisting */
				local_grey_on = 0;
				break;
			case DCC_GREY_ASK_EMBARGO:
				rcpt_st->flags |= RCPT_ST_GREY;
				wp->honor |= DCC_HONOR_LOGIT;
				wp->flags |= FG_GREY_LIST;
				break;
			case DCC_GREY_ASK_EMBARGO_END:
				rcpt_st->flags |= RCPT_ST_GREY_END;
				wp->honor |= DCC_HONOR_LOGIT;
				break;
			case DCC_GREY_ASK_PASS:
				break;
			case DCC_GREY_ASK_WHITE:
				rcpt_st->flags |= RCPT_ST_GREY_WHITE;
				break;
			}
		}
	}

	wp->log_pos_white_last = log_lseek(wp, SEEK_END);

	if (!dcc_white_unlock(wp->emsg, &cmn_wf, 0))
		dccifd_error_msg(wp, "%s", wp->emsg);
	dcc_wf_unlock(&cmn_wf);
}



/* ask a DCC server */
static void
ask_dcc(WORK *wp)
{
	int i;
	DCC_TGTS tgts;

	/* Talk to the DCC server and make the X-DCC header.
	 * If we have blacklist entries for it, then we'll tell the DCC
	 * server it is spam and say so in the X-DCC header.
	 * Note that a target count of 0 is a query. */
	if (query_only || (wp->flags & FG_MTA_QUERY)) {
		tgts = 0;
	} else if (wp->honor & DCC_HONOR_LOCAL_ISSPAM) {
		tgts = DCC_TGTS_TOO_MANY;
	} else if (wp->flags & FG_GREY_LIST) {
		tgts = wp->early_grey_tgts;
	} else {
		tgts = wp->tgts - wp->late_grey_tgts;
	}
	i = dcc_ask(wp->emsg, wp->dcc_ctxt, &wp->header, &wp->cks, &wp->honor,
		    tgts);
	if (i <= 0) {
		dccifd_error_msg(wp, "%s", wp->emsg);
		if (!i && try_extra_hard)
			job_done(wp, DCCIF_RESULT_TEMP, "temporary failure");
		/* after serious errors, act as if DCC server said not-spam */
		return;
	}

	/* if we are talking to a new server,
	 * remember to fix the X-DCC headers of the other contexts' */
	if (wp->xhdr_len != wp->header.start_len
	    || strncmp(wp->header.buf, wp->xhdr, wp->xhdr_len)) {
		if (dcc_clnt_debug)
			dccifd_trace_msg(wp, DCC_XHDR_START
				       "header changed from %s to %.*s",
				       wp->xhdr,
				       (int)wp->header.start_len,
				       wp->header.buf);
		wp->xhdr_len = dcc_xhdr_start(wp->xhdr, sizeof(wp->xhdr));
		wp->dcc_ctxt_sn = ++dcc_ctxt_sn;
	}

	/* log the header */
	log_write(wp, wp->header.buf, wp->header.used);
	LOG_CAPTION(wp, "\n");
}



static u_char
user_log_write(RCPT_ST *rcpt_st, const void *buf, u_int len)
{
	int result;

	if (rcpt_st->log_fd < 0)
		return 0;

	if (!len)
		len = strlen(buf);
	result = write(rcpt_st->log_fd, buf, len);
	if (result == (int)len)
		return 1;

	if (result < 0)
		dccifd_error_msg(rcpt_st->wp, "write(%s): %s",
			       DCC_NM2PATH(rcpt_st->log_nm),
			       strerror(errno));
	else
		dccifd_error_msg(rcpt_st->wp,
			       "write(%s)=%d instead of %d",
			       DCC_NM2PATH(rcpt_st->log_nm),
			       result, (int)len);
	close(rcpt_st->log_fd);
	rcpt_st->log_fd = -1;
	return 0;
}



static void
user_log_print(RCPT_ST *rcpt_st, const char *p, ...)
{
	char logbuf[MAXHOSTNAMELEN*2];
	va_list args;

	va_start(args, p);
	vsnprintf(logbuf, sizeof(logbuf), p, args);
	va_end(args);
	user_log_write(rcpt_st, logbuf, 0);
}



#define USER_LOG_CAPTION(rcpt_st, s) user_log_write((rcpt_st), (s), sizeof(s)-1)
#define USER_LOG_EOL(rcpt_st) USER_LOG_CAPTION((rcpt_st), "\n")

static void
user_log_ck_write(void *rcpt_st0, const void *buf, u_int len)
{
	user_log_write((RCPT_ST *)rcpt_st0, buf, len);
}



static void
user_log_block(RCPT_ST *rcpt_st,	/* copy from main log file to this */
	       off_t start,		/* starting here */
	       off_t stop)		/* and ending here */
{
	char buf[4096];
	off_t len;
	int result;

	if (start == -1 || stop == -1) {
		dccifd_error_msg(rcpt_st->wp, "bogus user_log_block position");
		job_exit(rcpt_st->wp);
	}

	if (-1 == lseek(rcpt_st->wp->log_fd2, start, SEEK_SET)) {
		dccifd_error_msg(rcpt_st->wp, "lseek(%s,%d,SEEK_SET): %s",
			       DCC_NM2PATH(rcpt_st->log_nm), (int)start,
			       strerror(errno));
		job_exit(rcpt_st->wp);
	}

	while ((len = stop - start) != 0) {
		if (len > (off_t)sizeof(buf))
			len = sizeof(buf);
		result = read(rcpt_st->wp->log_fd2, buf, len);
		if (result != len) {
			if (result < 0)
				dccifd_error_msg(rcpt_st->wp, "user_log_block"
					       " read(%s): %s",
					       DCC_NM2PATH(rcpt_st->wp->log_nm),
					       strerror(errno));
			else
				dccifd_error_msg(rcpt_st->wp, "user_log_block"
					       " read(%s)=%d instead of %d",
					       DCC_NM2PATH(rcpt_st->wp->log_nm),
					       result, (int)len);
			job_exit(rcpt_st->wp);
		}
		if (!user_log_write(rcpt_st, buf, len))
			job_exit(rcpt_st->wp);
		start += len;
	}
}



static void
user_log(RCPT_ST *rcpt_st)
{
	DCC_PATH rcpt_logdir;
	int log_mode;
	int i;

	snprintf(rcpt_logdir, sizeof(rcpt_logdir), "%s/log", rcpt_st->dir);
	i = dcc_logdir_ck(rcpt_st->wp->emsg, rcpt_logdir, &log_mode);
	if (i <= 0) {
		if (i < 0)
			dccifd_error_msg(rcpt_st->wp, "%s", rcpt_st->wp->emsg);
		return;
	}

	/* get an independent FD for the main log file that can be
	 * repositioned without affecting additional output to the main log. */
	if (rcpt_st->wp->log_fd2 < 0) {
		/* give up if things are already broken */
		if (rcpt_st->wp->log_fd2 != -1)
			return;
		/* Some systems don't synchronize the meta data among FDs for
		 * a file, causing the file second FD to appear to be
		 * truncated. */
		if (fsync(rcpt_st->wp->log_fd) < 0)
			dccifd_error_msg(rcpt_st->wp, "fsync(%s): %s",
				       DCC_NM2PATH(rcpt_st->wp->log_nm),
				       strerror(errno));
		rcpt_st->wp->log_fd2 = open(rcpt_st->wp->log_nm, O_RDONLY, 0);
		if (rcpt_st->wp->log_fd2 < 0) {
			dccifd_error_msg(rcpt_st->wp, "open(%s): %s",
				       DCC_NM2PATH(rcpt_st->wp->log_nm),
				       strerror(errno));
			rcpt_st->wp->log_fd2 = -2;
			return;
		}
	}

	/* create the user's log file */
	rcpt_st->log_fd = dcc_mkstemp(rcpt_st->log_nm, sizeof(rcpt_st->log_nm),
				      rcpt_logdir, DCC_FIN_LOG_PAT,
				      -1, log_mode & 0644);
	if (rcpt_st->log_fd < 0) {
		dccifd_error_msg(rcpt_st->wp, "dcc_mkstemp(%s): %s",
			       DCC_NM2PATH(rcpt_logdir), strerror(errno));
		return;
	}

	/* copy envelope before env_To line */
	user_log_block(rcpt_st,
		       0, rcpt_st->wp->log_pos_to_first);
	user_log_block(rcpt_st,		/* copy this env_To line */
		       rcpt_st->log_pos_to,
		       rcpt_st->fwd
		       ? rcpt_st->fwd->log_pos_to
		       : rcpt_st->wp->log_pos_to_end);
	user_log_block(rcpt_st,		/* copy the body of the message */
		       rcpt_st->wp->log_pos_to_end,
		       rcpt_st->wp->log_pos_white_first);
	user_log_block(rcpt_st,		/* copy white list error messages */
		       rcpt_st->log_pos_white,
		       rcpt_st->fwd
		       ? rcpt_st->fwd->log_pos_white
		       : rcpt_st->wp->log_pos_white_last);

	/* log the X-DCC header if it exists */
	USER_LOG_EOL(rcpt_st);
	if (rcpt_st->wp->header.buf[0] != '\0') {
		user_log_write(rcpt_st, rcpt_st->wp->header.buf,
			       rcpt_st->wp->header.used);
		USER_LOG_CAPTION(rcpt_st, "\n");
	}

	/* log the checksums and their counts */
	dcc_print_cks(rcpt_st, &rcpt_st->wp->cks, rcpt_st->wtgts,
		      user_log_ck_write);
	USER_LOG_EOL(rcpt_st);
	dcc_print_grey(rcpt_st, rcpt_st->grey_result, rcpt_st->embargo_num,
		       0, rcpt_st->env_to,
		       rcpt_st->pair_sum, rcpt_st->triple_sum,
		       user_log_ck_write);
}



static void
user_finish(WORK *wp, RCPT_ST *rcpt_st)
{
	if ((wp->flags & FG_GREY_LIST) && wp->deliver_tgts != 0) {
		user_log_print(rcpt_st, "result:"
			       " temporary greylist embargo #%d\n",
			       rcpt_st->embargo_num+1);
		return;
	}
	if (!(rcpt_st->flags & RCPT_ST_BLACK)) {
		if (rcpt_st->flags & RCPT_ST_GREY_END)
			USER_LOG_CAPTION(rcpt_st, "result: accept"
					 " after greylist embargo\n");
		else if (rcpt_st->flags & RCPT_ST_GREY_WHITE)
			USER_LOG_CAPTION(rcpt_st, "result: accept; "
					 " greylist whitelist\n");
		else
			USER_LOG_CAPTION(rcpt_st, "result: accept\n");
		return;
	}

	if (to_white_only
	    && !(rcpt_st->flags & RCPT_ST_OK2)) {
		USER_LOG_CAPTION(rcpt_st, "result: -W ignore and accept\n");
		return;
	}

	rcpt_st->flags |= RCPT_ST_DISCARD;

	/* if one of the other targets wants this message,
	 * try to remove this address from the list */
	if (wp->deliver_tgts != 0) {
		log_print(wp, "result: discard forced for %s\n",
			  rcpt_st->env_to);
		USER_LOG_CAPTION(rcpt_st,
				 "result: discard forced by"
				 " other target's white list\n");
		++totals.tgts_discarded;
		--wp->black_tgts;
		return;
	}

	USER_LOG_CAPTION(rcpt_st, "result: reject\n");
}



/* read lines of pairs if (env_To, username) values from the MTA */
static void
do_rcpts(WORK *wp)
{
	RCPT_ST *rcpt_st;
	char *rcpt, *user, c, *p;
	int rcpt_len;

	lock_work();
	wp->flags |= FG_WORK_LOCK;
	for (;;) {
		sock_read_line(wp);

		/* stop after the empty line */
		rcpt = wp->sbuf_out;
		if (*rcpt == '\0') {
			wp->sbuf_out = wp->sbuf_next_line;
			break;
		}

		++wp->tgts;

		/* bytes up to '\r' are the env_To value
		 * and bytes after are the local recipient */
		user = strchr(rcpt, '\r');
		if (!user) {
			rcpt_len = wp->sbuf_next_line - wp->sbuf_out;
		} else {
			*user++ = '\0';
			rcpt_len  = user - rcpt;
			if (*user == '\0'
			    || !strcmp(user, "..") || strchr(user, '/'))
				user = 0;
		}
		if (rcpt_len > ISZ(rcpt_st->env_to)) {
			dccifd_error_msg(wp, "recipient \"%s\" is too long",
					 rcpt);
			job_exit(wp);
		} else if (rcpt_len == 0) {
			dccifd_error_msg(wp, "null recipient");
			job_exit(wp);
		}

		rcpt_st = rcpt_st_free;
		if (rcpt_st) {
			rcpt_st_free = rcpt_st->fwd;
		} else {
			rcpt_st = dcc_malloc(sizeof(*rcpt_st));
			memset(rcpt_st, 0, sizeof(*rcpt_st));
			rcpt_st->log_fd = -1;
		}
		rcpt_st->fwd = 0;
		rcpt_st->wp = wp;
		rcpt_st->log_pos_white = 0;
		memset(rcpt_st->wtgts, 0, sizeof(rcpt_st->wtgts));
		rcpt_st->flags = 0;
		rcpt_st->dir[0] = '\0';
		rcpt_st->log_nm[0] = '\0';
		if (!wp->rcpt_st) {
			wp->rcpt_st = rcpt_st;
		} else {
			wp->rcpt_st_last->fwd = rcpt_st;
		}
		wp->rcpt_st_last = rcpt_st;

		memcpy(rcpt_st->env_to, rcpt, rcpt_len);

		if (user) {
			/* Convert ASCII upper to lower case.
			 * Be simplistic about international character
			 * sets and avoid locale and portability
			 * complications. */
			p = rcpt_st->user;
			do {
				c = *user++;
				if (c >= 'A' && c <= 'Z')
					c -= 'A' - 'a';
				*p++ = c;
			} while (c != '\0' && p < LAST(rcpt_st->user));
		} else {
			rcpt_st->user[0] = '\0';
		}

		if (userdirs && user)
			snprintf(rcpt_st->dir, sizeof(rcpt_st->dir), "%s%s",
				 userdirs, rcpt_st->user);
		else
			rcpt_st->dir[0] = '\0';

		wp->sbuf_out = wp->sbuf_next_line;
	}
	unlock_work();
	wp->flags &= ~FG_WORK_LOCK;
}



/* Get the next header.
 *	The line is copied to te temporary file and then null terminated
 *	in the buffer */
static u_char				/* 1=have one, 0=end of headers */
get_hdr(WORK *wp)
{
	int log_len, hdr_len, llen;
	char *bol, *eol;
	u_char at_eol;

	log_len = 0;
	hdr_len = 0;
	at_eol = 0;
	for (;;) {
		/* get another line of the header */
		bol = wp->sbuf_out + hdr_len;
		llen = wp->sbuf_in - bol;
		if (llen < 2) {
			if (hdr_len > DCC_HDR_CK_MAX) {
				if (log_len < hdr_len)
					tmp_write(wp, wp->sbuf_out+log_len,
						  hdr_len - log_len);
				hdr_len = DCC_HDR_CK_MAX;
				log_len = DCC_HDR_CK_MAX;
				if (llen > 0)
					memmove(wp->sbuf_out + DCC_HDR_CK_MAX,
						bol, llen);
				wp->sbuf_in = (wp->sbuf_out + DCC_HDR_CK_MAX
					       + llen);
			}
			if (!sock_read(wp)) {
				dccifd_error_msg(wp, "missing message body");
				job_exit(wp);
			}
			continue;
		}

		/* stop before the next line if it is not a continuation
		 * of the current header */
		if (at_eol
		    && *bol != ' '
		    && *bol != '\t') {
			if (log_len < hdr_len)
				tmp_write(wp, wp->sbuf_out+log_len,
					  hdr_len - log_len);
			wp->sbuf_out[hdr_len-1] = '\0';
			wp->sbuf_next_line = &wp->sbuf_out[hdr_len];
			return 1;
		}

		/* find the end of the next line */
		eol = memchr(bol, '\n', llen);
		if (eol) {
			hdr_len += ++eol - bol;
			if (hdr_len == 1
			    || (hdr_len == 2 && *wp->sbuf_out == '\r'))
				return 0;   /* quit at the end of headers */
			at_eol = 1;
		} else {
			hdr_len += llen;
			at_eol = 0;
		}
	}
}



static void
do_hdrs(WORK *wp)
{
	char *p, *p2;
	int i;

	for (;;) {
		/* stop at the separator between the body and headers */
		if (!get_hdr(wp))
			break;

#define GET_HDR_CK(h,t) {						    \
			if (!strncasecmp(wp->sbuf_out, h, sizeof(h)-1)) {   \
				dcc_get_cks(&wp->cks, DCC_CK_##t,	    \
					    &wp->sbuf_out[sizeof(h)-1], 1); \
				wp->flags |= FG_SEEN_HDR;		    \
				wp->sbuf_out = wp->sbuf_next_line;	    \
				continue;}}
		GET_HDR_CK(DCC_XHDR_TYPE_FROM":", FROM);
		GET_HDR_CK(DCC_XHDR_TYPE_MESSAGE_ID":", MESSAGE_ID);
#undef GET_HDR_CK

		if (!strncasecmp(wp->sbuf_out, "Return-Path:",
				 sizeof("Return-Path:")-1)) {
			if (wp->cks.sums[DCC_CK_ENV_FROM].type
			    == DCC_CK_INVALID) {
				BUFCPY(wp->env_from,
				       &wp->sbuf_out[sizeof("Return-Path:")-1]);
				dcc_get_cks(&wp->cks, DCC_CK_ENV_FROM,
					    wp->env_from, 1);
			}
			wp->flags |= FG_SEEN_HDR;
			wp->sbuf_out = wp->sbuf_next_line;
			continue;
		}

		/* notice UNIX From_ line */
		if (!(wp->flags & FG_SEEN_HDR)
		    && !strncmp(wp->sbuf_out, "From ", sizeof("From ")-1)) {
			p = &wp->sbuf_out[sizeof("From ")-1];
			p += strspn(p, " ");
			p2 = strchr(p, ' ');
			if (p2 != 0) {
				if (wp->cks.sums[DCC_CK_ENV_FROM].type
				    == DCC_CK_INVALID) {
					if (p2 > p+sizeof(wp->env_from))
					    p2 = p+sizeof(wp->env_from);
					memcpy(wp->env_from, p, p2-p);
					wp->env_from[p2-p] = '\0';
				}
				wp->flags |= FG_SEEN_HDR;
				wp->sbuf_out = wp->sbuf_next_line;
				continue;
			}
		}

		if (!strncasecmp(wp->sbuf_out, DCC_XHDR_TYPE_RECEIVED":",
				 sizeof(DCC_XHDR_TYPE_RECEIVED":")-1)) {
			p = &wp->sbuf_out[sizeof(DCC_XHDR_TYPE_RECEIVED":")-1];

			/* compute checksum of the last Received: header */
			dcc_get_cks(&wp->cks, DCC_CK_RECEIVED, p, 1);

			/* pick IP address out of first Received: header */
			if ((wp->flags & FG_CK_RCVD)
			    && wp->cks.sums[DCC_CK_IP].type == DCC_CK_INVALID
			    && 0 != (p2 = strchr(p, '('))
			    && 0 != (p = strchr(++p2, '['))
			    && (i = strspn(++p, ".:abcdefABCDEF0123456789")) > 6
			    && i < INET6_ADDRSTRLEN
			    && p[i] == ']') {
				p[i] = '\0';
				if (dcc_get_str_ip_ck(&wp->cks, p,
						      &wp->clnt_addr)) {
					if (!DCC_INET_NTOP(AF_INET6,
							&wp->clnt_addr,
							wp->clnt_str,
							sizeof(wp->clnt_str)))
					    strcpy(wp->clnt_str, p);

					i = p-p2-1;
					if (i > ISZ(wp->clnt_name)-1)
					    i = ISZ(wp->clnt_name)-1;
					while (i > 0
					       && (p2[i-1] == ' '
						   || p2[i-1] == '\t'))
					    --i;
					if (i > 0) {
					    memcpy(wp->clnt_name, p2, i);
					    wp->clnt_name[i] = '\0';
					} else {
					    snprintf(wp->clnt_name,
						     sizeof(wp->clnt_name),
						     "[%s]", p);
					}
				}
			}
			wp->flags &= ~FG_CK_RCVD;
			wp->flags |= FG_SEEN_HDR;
			wp->sbuf_out = wp->sbuf_next_line;
			continue;
		}

		/* Notice MIME multipart boundary definitions */
		dcc_ck_mime_hdr(&wp->cks, wp->sbuf_out, 0);

		if (dcc_ck_get_sub(&wp->cks, wp->sbuf_out, 0))
			wp->flags |= FG_SEEN_HDR;

		/* notice any sort of header */
		if (!(wp->flags & FG_SEEN_HDR)) {
			for (p = wp->sbuf_out; ; ++p) {
				if (*p == ':') {
					wp->flags |= FG_SEEN_HDR;
					break;
				}
				if (*p <= ' ' || *p >= 0x7f)
					break;
			}
		}

		wp->sbuf_out = wp->sbuf_next_line;
	}

	/* Create a checksum for a null Message-ID header if there
	 * was no Message-ID header.  */
	if (wp->cks.sums[DCC_CK_MESSAGE_ID].type != DCC_CK_MESSAGE_ID)
		dcc_get_cks(&wp->cks, DCC_CK_MESSAGE_ID, "", 0);
}



/* start a new connection to an SMTP client */
static void *
job_start(void *wp0)
{
	WORK *wp = wp0;
	sigset_t sigsoff;
	RCPT_ST *rcpt_st;
	char *p, *p2;
	char buf[1024];
	u_char need_eol;
	int buflen, i;

	i = pthread_sigmask(SIG_BLOCK, &sigsoff, 0);
	if (i)
		dcc_logbad(EX_SOFTWARE, "pthread_sigmask(): %s",
			   strerror(i));
	log_start(wp);

	/* get any options */
	sock_read_line(wp);
	while (*wp->sbuf_out != '\0') {
		if (sock_ck_str(wp, DCCIF_OPT_SPAM,
				sizeof(DCCIF_OPT_SPAM)-1)) {
			wp->flags |= FG_MTA_ISSPAM;
			continue;
		}
		if (sock_ck_str(wp, DCCIF_OPT_BODY,
				sizeof(DCCIF_OPT_BODY)-1)) {
			wp->flags |= FG_MTA_BODY;
			continue;
		}
		if (sock_ck_str(wp, DCCIF_OPT_HEADER,
				sizeof(DCCIF_OPT_HEADER)-1)) {
			wp->flags |= FG_MTA_HEADER;
			continue;
		}
		if (sock_ck_str(wp, DCCIF_OPT_QUERY,
				sizeof(DCCIF_OPT_QUERY)-1)) {
			wp->flags |= FG_MTA_QUERY;
			continue;
		}

		dccifd_error_msg(wp, "unrecognized option value: \"%s\"",
				 wp->sbuf_out);
		job_exit(wp);
	}
	if ((wp->flags & FG_MTA_ISSPAM)
	    && (wp->flags & FG_MTA_QUERY)) {
		dccifd_error_msg(wp, DCCIF_OPT_SPAM" and "DCCIF_OPT_QUERY
				 " are incompatible");
		wp->flags &= ~FG_MTA_ISSPAM;
	}
	wp->sbuf_out = wp->sbuf_next_line;

	/* deal with the SMTP client IP address and host name */
	i = sock_read_line(wp);
	if (i == 0) {
		wp->clnt_name[0] = '\0';
		wp->clnt_str[0] = '\0';
		wp->flags |= FG_CK_RCVD;    /* try first Received header */
	} else {
		/* the host name follows the IP address */
		p = strchr(wp->sbuf_out, '\r');
		if (p) {
			*p++ = '\0';
			BUFCPY(wp->clnt_name, p);
		}
		/* convert IP address to an IPv6 number */
		if (!dcc_get_str_ip_ck(&wp->cks, wp->sbuf_out,
				       &wp->clnt_addr)) {
			dccifd_error_msg(wp, "unrecognized IP address \"%s\"",
					 wp->sbuf_out);
		} else {
			/* convert IPv6 address to a canonical string */
			if (!DCC_INET_NTOP(AF_INET6, &wp->clnt_addr,
					   wp->clnt_str,
					   sizeof(wp->clnt_str))) {
				dccifd_error_msg(wp,
						 "inet_ntop(addr of %s) failed",
						 wp->sbuf_out);
				strcpy(wp->clnt_str, "(inet_ntop() ?)");
			}
		}
	}
	wp->sbuf_out = wp->sbuf_next_line;

	/* deal with the HELO value */
	i = sock_read_line(wp);
	if (i > DCC_HELO_MAX-1) {
		i = DCC_HELO_MAX-1;
		strcpy(&wp->sbuf_out[DCC_HELO_MAX-ISZ(DCC_HELO_CONT)],
		       DCC_HELO_CONT);
	}
	memcpy(wp->helo, wp->sbuf_out, i+1);
	wp->sbuf_out = wp->sbuf_next_line;

	/* get the envelope mail_from value */
	i = sock_read_line(wp);
	if (i > ISZ(wp->env_from)-1) {
		i = ISZ(wp->env_from)-1;
		wp->sbuf_out[ISZ(wp->env_from)] = '\0';
	}
	memcpy(wp->env_from, wp->sbuf_out, i+1);
	wp->sbuf_out = wp->sbuf_next_line;

	/* get the recipients */
	do_rcpts(wp);

	/* We must make a copy of the message in a temporary file
	 * if the MTA wants a copy with the X-DCC header added
	 * Copy the headers to a temporary file because the official
	 * log file needs the SMTP client IP address and envelope information
	 * before the header lines.  The log file needs all of the header
	 * lines including stray X-DCC lines, but those lines must be removed
	 * from the output file. */
	wp->tmp_fd = dcc_mkstemp(wp->tmp_nm, sizeof(wp->tmp_nm),
				 tmpdir, "/dccifd.", 0, 0);
	if (wp->tmp_fd < 0)
		dcc_logbad(EX_IOERR, "open(%s): %s",
			   wp->tmp_nm, strerror(errno));
	if (0 > unlink(wp->tmp_nm))
		dcc_logbad(EX_IOERR, "unlink(%s): %s",
			   wp->tmp_nm, strerror(errno));

	/* get ready for the body checksums before the headers so that
	 * we can notice the MIME separator */
	dcc_ck_body_init(&wp->cks);

	/* open the connection to the nearest DCC server
	 * and figure out our X-DCC header */
	if (wp->dcc_ctxt_sn != dcc_ctxt_sn) {
		wp->dcc_ctxt_sn = dcc_ctxt_sn;
		wp->dcc_ctxt = dcc_clnt_init(wp->emsg, wp->dcc_ctxt,
					     0, 0, 1, 1);
		if (!wp->dcc_ctxt) {
			/* failed to create context */
			dccifd_error_msg(wp, "%s", wp->emsg);
			wp->dcc_ctxt_sn = 0;
			job_done(wp, DCCIF_RESULT_TEMP, "temporary failure");
		}
		wp->xhdr_len = dcc_xhdr_start(wp->xhdr, sizeof(wp->xhdr));
	}

	/* get the headers */
	do_hdrs(wp);

	/* log IP address, env_From, and env_To values now
	 * that we (may) have collected from the headers */
	if (wp->cks.sums[DCC_CK_IP].type != DCC_CK_INVALID) {
		log_print(wp, DCC_XHDR_TYPE_IP": %s %s\n",
			  wp->clnt_name, wp->clnt_str);
	}
	if (wp->helo[0] != '\0')
		log_print(wp, "HELO: %s\n", wp->helo);
	dcc_ck_get_sub(&wp->cks, "helo", wp->helo);
	if (wp->env_from[0] != '\0') {
		LOG_CAPTION(wp, DCC_XHDR_TYPE_ENV_FROM": ");
		log_write(wp, wp->env_from, 0);
		dcc_get_cks(&wp->cks, DCC_CK_ENV_FROM, wp->env_from, 1);

		LOG_CAPTION(wp, "  mail_host=");
		p = strchr(wp->env_from, '@');
		if (p) {
			p2 = strchr(p, '>');
			if (p2)
				*p2 = '\0';
			log_write(wp, p, 0);
			dcc_ck_get_sub(&wp->cks, "mail_host", p);
		}
		LOG_EOL(wp);
	}
	wp->log_pos_to_first = log_lseek(wp, SEEK_END);
	wp->log_pos_to_end = wp->log_pos_to_first;
	for (rcpt_st = wp->rcpt_st; rcpt_st; rcpt_st = rcpt_st->fwd) {
		rcpt_st->log_pos_to = wp->log_pos_to_end;
		LOG_CAPTION(wp, DCC_XHDR_TYPE_ENV_TO": ");
		log_write(wp, rcpt_st->env_to, 0);
		LOG_CAPTION(wp, "  addr=");
		log_write(wp, rcpt_st->user, 0);
		LOG_CAPTION(wp, "  dir=");
		log_write(wp, rcpt_st->dir, 0);
		LOG_EOL(wp);
		wp->log_pos_to_end = log_lseek(wp, SEEK_END);
	}

	/* log the blank line between the header and the body */
	LOG_EOL(wp);

	/* copy headers from tmp file to log file */
	if (wp->log_fd >= 0) {
		if (0 > lseek(wp->tmp_fd, 0, SEEK_SET)) {
			dccifd_error_msg(wp, "rewind %s: %s",
					 wp->tmp_nm, strerror(errno));
			job_exit(wp);
		}
		for (;;) {
			buflen = tmp_read(wp, buf, sizeof(buf));
			if (buflen <= 0)
				break;
			log_write(wp, buf, buflen);
		}
	}
	if (!(wp->flags & FG_MTA_BODY)) {
		if (0 > close(wp->tmp_fd))
			dccifd_error_msg(wp, "close(%s): %s",
					 wp->tmp_nm, strerror(errno));
		wp->tmp_fd = -1;
	}

	/* collect the body */
	for (;;) {
		buflen = wp->sbuf_in - wp->sbuf_out;
		if (buflen <= 0) {
			if (!sock_read(wp))
				break;
			buflen = wp->sbuf_in - wp->sbuf_out;
		}

		/* Log the body block
		 * avoid filling the disk by truncating the body */
		if (wp->log_fd >= 0) {
			i = MAX_LOG_SIZE - wp->log_size;
			if (i >= 0) {
				if (i >= buflen) {
					log_write(wp, wp->sbuf_out, buflen);
				} else {
					p = wp->sbuf_out+i;
					while (--p > wp->sbuf_out) {
					    if (*p == '\n') {
						i = p-wp->sbuf_out+1;
						break;
					    }
					}
					log_write(wp, wp->sbuf_out, i);
					if (wp->sbuf_out[i-1] != '\n')
					    LOG_EOL(wp);
					LOG_CAPTION(wp, DCC_LOG_TRN_MSG_CR);
					wp->log_size = MAX_LOG_SIZE+1;
				}
			}
		}

		if (wp->flags & FG_MTA_BODY)
			tmp_write(wp, wp->sbuf_out, buflen);

		dcc_ck_body(&wp->cks, wp->sbuf_out, buflen);
		wp->sbuf_out = wp->sbuf_in;
	}
	dcc_ck_body_fin(&wp->cks);

	LOG_CAPTION(wp, DCC_LOG_MSG_SEP);

	/* check the whitelists */
	ask_white(wp);

	LOG_CAPTION(wp, "\n");
	wp->header.buf[0] = '\0';
	wp->header.used = 0;
	if (wp->tgts != wp->white_tgts || wp->tgts == 0) {
		/* report to the DCC and add our header only if allowed */
		ask_dcc(wp);
	} else {
		/* if not ok to ask, remove our X-DCC header */
		wp->cks.flags &= ~DCC_CKS_HAVE_SUM;
		/* log it if the target count is high enough */
		dcc_honor_log_cnts(&wp->honor, &wp->cks,
				   ((wp->honor & DCC_HONOR_LOCAL_ISSPAM)
				    ? DCC_TGTS_TOO_MANY : wp->tgts));
	}

	totals.tgts += wp->tgts;

	/* log the checksums, DCC server counts and global whitelist values */
	dcc_print_cks(wp, &wp->cks, wp->wtgts, log_ck_write);
	LOG_EOL(wp);

	/* mark recipients who won't receive it */
	need_eol = 0;
	for (rcpt_st = wp->rcpt_st; rcpt_st; rcpt_st = rcpt_st->fwd) {
		dcc_print_grey(wp, rcpt_st->grey_result, rcpt_st->embargo_num,
			       &need_eol,
			       rcpt_st->env_to,
			       rcpt_st->pair_sum,
			       rcpt_st->triple_sum,
			       log_ck_write);

		/* this recipient will receive it */
		if (rcpt_st->flags & RCPT_ST_WHITE)
			continue;

		/* recall whether the DCC server said it is spam */
		if (wp->honor & DCC_HONOR_SRVR_ISSPAM)
			rcpt_st->flags |= RCPT_ST_BLACK;

		if (rcpt_st->flags & RCPT_ST_BLACK) {
			if (wp->flags & FG_MTA_ISSPAM) {
				++wp->black_tgts;
				continue;
			}
			if (!to_white_only
			    || (rcpt_st->flags & RCPT_ST_OK2)) {
				++wp->black_tgts;
				continue;
			}
		}

		++wp->deliver_tgts;
	}
	if (need_eol)
		LOG_EOL(wp);

	/* create individual log files
	 * and trim target list if necessary */
	for (rcpt_st = wp->rcpt_st; rcpt_st; rcpt_st = rcpt_st->fwd) {
		if (!(wp->honor & DCC_HONOR_LOGIT))
			continue;	/* not if not needed */

		if (dcc_have_logdir
		    && rcpt_st->dir[0] != '\0')
			user_log(rcpt_st);

		user_finish(wp, rcpt_st);

		if (rcpt_st->log_fd >= 0) {
			if (0 > close(rcpt_st->log_fd))
				dccifd_error_msg(wp, "close(%s): %s",
						 rcpt_st->log_nm,
						 strerror(errno));
			rcpt_st->log_fd = -1;
		}
	}

	/* deliver it if all remaining targets want it */
	totals.tgts_rejected += wp->black_tgts;
	if (wp->black_tgts != 0
	    || (wp->tgts == 0 && (wp->honor & DCC_HONOR_SRVR_ISSPAM)))
		job_done(wp, DCCIF_RESULT_REJECT, "reject");
	if (wp->flags & FG_GREY_LIST) {
		totals.tgts_embargoed += wp->tgts;
		job_done(wp, DCCIF_RESULT_GREY, "embargo");
	}
	if (wp->deliver_tgts == wp->tgts)
		job_done(wp, DCCIF_RESULT_OK, "accept");
	job_done(wp, DCCIF_RESULT_SOME, "accept for some");
}
