%define name dcc-dccd
%define version 1.1.20
%define release 1

Name: %{name}
Summary: Distributed Checksum Clearinghouse, anti-spam tool
Version: %{version}
Release: %{release}
Copyright: BSD-like
Group: Networking/Mail
Source: http://www.dcc-servers.net/dcc/source/%{name}-%{version}.tar.Z
Buildroot: %{_tmppath}/%{name}-%{version}-root
Prereq: /sbin/chkconfig /usr/sbin/useradd
BuildRequires: sendmail-devel >= 8.12
Requires: sendmail >= 8.12, sendmail-cf >= 8.12

%description
Distributed Checksum Clearinghouse or DCC is a cooperative, distributed
system intended to detect "bulk" mail or mail sent to many people.
It allows individuals receiving a single mail message to determine that
many other people have been sent essentially identical copies of the
message and so reject the message.  It can identify some unsolicited bulk
mail using "spam traps" and other detectors, but that is not its focus.

The DCC can be viewed as a tool for end users to enforce their right to
"opt-in" to streams of bulk mail by refusing all bulk mail except from
sources in a "white list."  White lists are generally the responsibility
of DCC clients, since only they know which bulk mail they solicited.

NB to use DCC to reject SPAM you need to configure /var/dcc/dcc_conf
and either use procmail or sendmail to feed the messages to DCC

%package cgi
Summary: cgi-scripts for managing mail delivery on a DCC enabled server
Group: Networking/Mail
Requires: %{name} >= 1.1.2

%description cgi
Example set of cgi-scripts to allow users to point-and-click manage their
own DCC whitelists and thus what is delivered to them.  Allows overriding
of site level lists.  The scripts give controlled access to the whitelists
which are otherwise in protected directory space (owned by dcc).

NB these scripts need configured after installation

%prep
[ $RPM_BUILD_ROOT != "/" ] && rm -rf $RPM_BUILD_ROOT
%setup

%build
: "We expect milter to be installed by the sendmail-devel rpm in /usr/lib/libmilter.a"
: "this uses the same tests as for FreeBSD 4.6 in the configure script"
: "NB the RedHat 7.3 libmilter calls up strlcpy rather than sm_strlcpy and does not link"
%configure --with-sendmail --bindir='$(bindir)' --mandir='$(mandir)' --libexecdir='$(sbindir)'  --prefix=/var/dcc --program-prefix=/usr --with-uid=dcc --with-db-memory=32
perl -p -i.orig -e "s:\".*\":\"/usr/sbin\": if m/define\s+DCC_LIBEXECDIR/ ;" include/dcc_config.h
make

%install
MANOWN=`id -u` MANGRP=`id -g`  INSTALL="install -C"  export MANOWN MANGRP INSTALL
BINOWN=$MANOWN BINGRP=$MANGRP export BINOWN BINGRP
DCC_PROTO_HOMEDIR=$RPM_BUILD_ROOT/var/dcc/ export DCC_PROTO_HOMEDIR
DCC_CGIBINDIR=$RPM_BUILD_ROOT/var/dcc/cgi-bin export DCC_CGIBINDIR
DCC_SUID=$BINOWN DCC_OWN=$BINOWN DCC_GRP=$BINGRP export DCC_SUID DCC_OWN DCC_GRP
mkdir -p $RPM_BUILD_ROOT/usr/share/sendmail-cf/feature $DCC_CGIBINDIR
mkdir -p $RPM_BUILD_ROOT/etc/{init.d,cron.daily} $RPM_BUILD_ROOT/{var/{run/dcc,dcc/{log,userdirs/{local,esmtp,cyrus,procmail}}},usr/bin}
%makeinstall
cp misc/dcc.m4 misc/dccdnsbl.m4 $RPM_BUILD_ROOT/usr/share/sendmail-cf/feature
mv $RPM_BUILD_ROOT/usr/sbin/cron-dccd $RPM_BUILD_ROOT/etc/cron.daily/dccd

# There already is a logger program which takes the required arguments on Linux

rm -f $RPM_BUILD_ROOT/usr/sbin/logger

# We are putting these in the docs directory
rm -f $RPM_BUILD_ROOT/usr/sbin/hackmc
rm -f $RPM_BUILD_ROOT/usr/sbin/na-spam
rm -f $RPM_BUILD_ROOT/usr/sbin/ng-spam

mv $RPM_BUILD_ROOT/usr/sbin/rcDCC $RPM_BUILD_ROOT/etc/init.d/dccd
chmod 755 $RPM_BUILD_ROOT/etc/init.d/dccd

# Set some initial logging, but no rejections
perl -p -i -e "s/BRAND=\$/BRAND=RPMDEFAULT/ ; s/DCCM_LOG_AT=\$/\$&10/ ; " \
	$RPM_BUILD_ROOT/var/dcc/dcc_conf

umask 077
cat > $RPM_BUILD_ROOT/var/dcc/flod <<EOF
# hostname,port     rem-id [passwd-id] [out-opts] [in-opts]

# this will not work
# nonesuch.example.com,-	32767 100
EOF

%clean
[ $RPM_BUILD_ROOT != "/" ] && rm -rf $RPM_BUILD_ROOT
make clean

%pre
/usr/sbin/useradd -d /var/dcc -r dcc >/dev/null 2>&1 || :

%postun
if [ $1 = 0 ]; then
        /usr/sbin/userdel -r dcc > /dev/null 2>&1 || :
fi

%post
/sbin/chkconfig --add dccd || :
/sbin/chkconfig --level 016 dccd off || :
/usr/bin/cdcc info > /var/dcc/map.txt || :

%post cgi
echo The scripts need configured and added into your web configuration.
echo see /var/dcc/cgi-bin/README


%preun
if [ $1 = 0 ]; then
	#
	# This is remove rather than update
	#
	/sbin/chkconfig --del dccd || :
	/etc/init.d/dccd stop || :
fi

%files
%defattr(-,dcc,root)

%dir	/var/dcc
%dir	/var/dcc/log
%dir	/var/dcc/userdirs/local
%dir	/var/dcc/userdirs/cyrus
%dir	/var/dcc/userdirs/procmail
%dir	/var/dcc/userdirs/esmtp
%dir	/var/run/dcc
%doc CHANGES FAQ.html FAQ.txt INSTALL.html INSTALL.txt LICENSE cdcc.html dbclean.html dblist.html dccd.html dcc.html dccm.html dccproc.html dccsight.html homedir/flod homedir/ids homedir/map.txt homedir/README misc/dcc.m4 misc/dccdnsbl.m4 misc/hackmc misc/na-spam misc/ng-spam 
%config(noreplace)	/var/dcc/dcc_conf
%config(noreplace)	/var/dcc/whiteclnt
%config(noreplace)	/var/dcc/whitecommon
%config(noreplace)	/var/dcc/whitelist
%config(noreplace)	/var/dcc/ids
%config(noreplace)	/var/dcc/flod
%config(noreplace)	/var/dcc/dcc_db
%config(noreplace)	/var/dcc/dcc_db.hash
%config(noreplace)	/var/dcc/map
/var/dcc/map.txt
/etc/init.d/dccd
/etc/cron.daily/dccd
/usr/bin/cdcc
/usr/bin/dccproc
/usr/sbin/dbclean
/usr/sbin/dblist
/usr/sbin/dccd
/usr/sbin/dccm
/usr/sbin/dccsight
/usr/sbin/refeed
/usr/sbin/start-dccd
/usr/sbin/start-dccm
/usr/sbin/stop-dccd
/usr/sbin/wlist
/usr/share/sendmail-cf/feature/dcc.m4
/usr/share/sendmail-cf/feature/dccdnsbl.m4

%{_mandir}/man8/*

%files cgi
%defattr(-,dcc,root)
%dir	/var/dcc/cgi-bin
/var/dcc/cgi-bin/chgpasswd
/var/dcc/cgi-bin/common
/var/dcc/cgi-bin/edit-whiteclnt
/var/dcc/cgi-bin/http2https
/var/dcc/cgi-bin/list-log
/var/dcc/cgi-bin/list-msg
/var/dcc/cgi-bin/README
/var/dcc/cgi-bin/webuser-notify

%changelog
* Wed Jul 3 2002 Andrew Macpherson <andrew@oa5.com>
- Version 1.1.4
- Checkpoint against changes and patches
- Add sub-package for the cgi-management tools
- Delete  hooks to pre-build databases, this is now done in makeinstall.
- 
* Sat May 25 2002 Andrew Macpherson <andrew@oa5.com>
- Version 1.1.2.a
- Patch free version that assumes Vernon will make
- changes to Man page autoconf, and homedir/Makefile.in
- 
* Sat May 25 2002 Andrew Macpherson <andrew@oa5.com>
- Version 1.1.2
- Big changes to the sources by Vernon Schryver mean that
- patches can be largely eliminated.  Still need to
- replace the man page installation code in the original
- tar, but the reworked FreBSD autoconf code works for me, and 
- the homedir/Makefile patch is a preview of code proposed
- by Vernon for the next release
- 
- Moving installation location of 3 scripts into the doc
- directory, as they are not really common-use.
- 
- Setup as client-only using the publicly-listed anonymous
- servers, and include the OA5 servers as well, though they
- are not for public listing
- 
- Do not start dccd/dccm on install
- 
- Moved messing with dcc_conf settings to pre-packaging vs install script
- 
- Created skeleton userdirs for per recipient logging
- 
- rpm -e , droped deleting databases in favour of removing
- Home directory on complete remove vs upgrade
- 
* Mon May 15 2002 Andrew Macpherson <andrew@oa5.com>
- Version 1.0.47
- Basic setup, fighting the appalling autoconf (I hate that package)
- to try to get things to install in sensible places
- gave up and overwrote makefiles
- 
- Make an initial setup that will let people run 
- stand alone.  Considered and rejected configuring to use 
- anonymous clients against publicly available servers
- 
- Patch the rc script so that chkconfig will operate
